"""Command line comleter helper for python."""

from __future__ import annotations

from os import listdir
from os.path import split
from pathlib import Path
from pwd import getpwall
from rlcompleter import Completer
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


__date__ = "2024/08/05 23:41:18 hoel"
__author__ = "Sunjoong LEE <sunjoong@gmail.com>"
__copyright__ = """Copyright © 2006 by Sunjoong LEE
Copyright © 2020, 2022, 2024 by Berthold Höllmann"""
__credits__ = ["Sunjoong LEE", "Berthold Höllmann"]
__maintainer__ = "Berthold Höllmann"
__email__ = "berhoel@gmail.com"


class IrlCompleter(Completer):
    """Commandline completer."""

    def __init__(self, namespace: dict[str, Any] | None = None) -> None:
        """Intialize class instance."""
        super().__init__(namespace=namespace)
        self.matches = [""]

    def complete(self, text: str, state: int) -> str | None:
        """Complete current command."""
        if text == "":
            # you could replace '    ' to \t if you indent via tab
            return ["    ", None][state]
        if text.count("'") == 1:
            if not state:
                self.file_matches(text, "'")
            try:
                return self.matches[state]
            except IndexError:
                return None
        if text.count('"') == 1:
            if not state:
                self.file_matches(text, '"')
            try:
                return self.matches[state]
            except IndexError:
                return None
        else:
            return Completer.complete(self, text, state)

    def file_matches(self, text: str, mark: str) -> None:
        """Check if input matche a file."""
        if "~" in text:
            if "/" in text:
                text = f'{mark}{Path(text[text.find("~"):]).expanduser()}'
            else:
                self.user_matches(text, mark)
                return

        text1 = text[1:]
        delim = "/"

        if not text1:
            directory = ""
        elif text1 == ".":
            directory = "."
        elif text1 == "..":
            directory = ".."
        elif text1 == "/":
            directory = "/"
            delim = ""
        elif text1.endswith("/"):
            directory = text1[:-1]
            delim = text1[len(directory) :]
        else:
            directory, partial = split(text1)
            delim = text1[len(directory) :][: -len(partial)]

        if directory:
            listing = [f"{mark}{directory}{delim}{x}" for x in listdir(directory)]
        else:
            listing = [f"{mark}{x}" for x in listdir(".")]

        n = len(text)
        self.matches = [x for x in listing if x[:n] == text]

    def user_matches(self, text: str, mark: str) -> None:
        """Upate `self.matches` according to input."""
        self.matches = [
            f"{mark}~{x.pw_name}" for x in getpwall() if x.pw_name.startswith(text)
        ]



"""Handle command line history."""

from __future__ import annotations

from itertools import count as icount
from pathlib import Path
from readline import (
    clear_history,
    get_completer_delims,
    get_current_history_length,
    get_history_item,
    parse_and_bind,
    read_history_file,
    set_completer,
    set_completer_delims,
    set_history_length,
    write_history_file,
)
import sys
from tempfile import mkstemp

# Local library imports.
from .irl_completer import IrlCompleter

__date__ = "2024/08/05 23:45:50 hoel"
__author__ = "Sunjoong LEE <sunjoong@gmail.com>"
__copyright__ = """Copyright © 2006 by Sunjoong LEE
Copyright © 2020, 2022, 2024 Berthold Höllmann"""
__credits__ = ["Sunjoong LEE", "Berthold Höllmann"]
__maintainer__ = "Berthold Höllmann"
__email__ = "berhoel@gmail.com"

HISTORY_PATH = Path.home() / f".pyhistory{sys.version_info.major}"
HISTORY_LENGTH = 100


class History:
    """Manage command history."""

    def __init__(self) -> None:
        """Intialize class instance."""
        self.recall()
        set_history_length(HISTORY_LENGTH)

        parse_and_bind("tab: complete")
        delims = get_completer_delims()
        set_completer_delims(delims)
        set_completer(IrlCompleter().complete)

    def __repr__(self) -> str:
        """Print out current history information."""
        length = get_current_history_length()
        if length > 1:
            try:
                return "\n".join(get_history_item(i) for i in range(1, length))
            except UnicodeDecodeError:
                return b"\n".join(
                    get_history_item(i).encode() for i in range(1, length)
                ).decode()

        else:
            return ""

    def __call__(self) -> None:
        """Print out current history information with line number."""
        length = get_current_history_length()
        if length > 1:
            kount = icount(1)
            try:
                for command in [get_history_item(i) for i in range(1, length)]:
                    print(f"{next(kount)}\t{command}")  # noqa:T201
            except UnicodeDecodeError:
                print(b"{next(kount)}\t{command}")  # noqa:T201

    @staticmethod
    def save(filename: Path, pos: int | None = None, end: int | None = None) -> None:
        """Write history number from pos to end into filename file."""
        length = get_current_history_length()
        if length > 1:
            if not pos:
                pos = 1
            elif pos >= length - 1:
                pos = length - 1
            elif pos < 1:
                pos = length + pos - 1
            if not end or end >= length:
                end = length
            end = length + end if end < 0 else end + 1

            with filename.open("w") as f_p:
                if pos < end:
                    for i in range(pos, end):
                        f_p.write(f"{get_history_item(i)}\n")
                else:
                    f_p.write(f"{get_history_item(pos)}\n")

    @staticmethod
    def clear() -> None:
        """Save the current history and clear it."""
        write_history_file(HISTORY_PATH)
        clear_history()

    @staticmethod
    def recall(history_path: Path = HISTORY_PATH) -> None:
        """Clear the current history and recall it from saved."""
        clear_history()
        if history_path.exists():
            read_history_file(history_path)

    @staticmethod
    def execute(pos: int, end: int | None = None) -> None:
        """Execute history number from pos to end."""
        length = get_current_history_length()
        if length > 1:
            if pos >= length - 1:
                pos = length - 1
            elif pos < 1:
                pos = length + pos - 1
            if not end:
                end = pos + 1
            elif end >= length:
                end = length
            end = length + end if end < 0 else end + 1

            filename = Path(mkstemp()[1])
            with filename.open("w") as f_p:
                for i in range(pos, end):
                    f_p.write(f"{get_history_item(i)}\n")
            try:
                with filename.open("rb") as f_p:
                    # exec previously saved commands
                    exec(compile(f_p.read(), filename, "exec"), locals())  # noqa:S102
                read_history_file(filename)
            except Exception:  # noqa:BLE001,S110 (ignore all errors)
                pass
            filename.unlink()



"""Save command line history and provide a command line completer for python."""

from __future__ import annotations

import atexit
from readline import set_pre_input_hook, write_history_file
from typing import TYPE_CHECKING

from .history import HISTORY_PATH, History

if TYPE_CHECKING:
    from pathlib import Path

__date__ = "2024/08/05 23:15:09 hoel"
__author__ = "Sunjoong LEE <sunjoong@gmail.com>"
__copyright__ = """Copyright © 2006 by Sunjoong LEE
Copyright © 2020, 2022, 2024 Berthold Höllmann"""
__credits__ = ["Sunjoong LEE", "Berthold Höllmann"]
__maintainer__ = "Berthold Höllmann"
__email__ = "berhoel@gmail.com"


def save_history(history_path: Path | None = None) -> None:
    """Save command history at program exit."""
    if history_path is None:
        history_path = HISTORY_PATH
    write_history_file(history_path)


atexit.register(save_history)


def hook() -> None:
    """Clean up before rl completer is executed."""
    set_pre_input_hook()
    del locals()["History"]
    del locals()["__file__"]


set_pre_input_hook(hook)

locals()["__builtins__"]["history"] = History()



"""Initialization for berhoel python namespace packages."""

__date__ = "2024/08/05 23:15:28 hoel"
__author__ = "Berthold Höllmann"
__copyright__ = "Copyright © 2020, 2024 by Berthold Höllmann"
__credits__ = ["Berthold Höllmann"]
__maintainer__ = "Berthold Höllmann"
__email__ = "berhoel@gmail.com"


__path__ = __import__("pkgutil").extend_path(__path__, __name__)
